#include "xmame.h"
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>

typedef enum
{
	kPlainFile,
	kZippedFile
} eFileType;

typedef struct
{
	FILE		*file;
	unsigned char	*data;
	int		offset;
	int		length;
 	eFileType	type;
} FakeFileHandle;

int load_zipped_file (const char *zipfile,const char *filename,
	unsigned char **buf, int *length);

/*
 * file handling routines
 *
 * gamename holds the driver name, filename is only used for ROMs and samples.
 * if 'write' is not 0, the file is opened for write. Otherwise it is opened
 * for read.
 */


/*
 * check if roms/samples for a game exist at all
 * return 1 on success, otherwise 0
 */
int osd_faccess(const char *newfilename, int filetype)
{
#ifndef MESS
	char name[256]; /* FIXME buffer overrun !! */
#endif
	if (access(newfilename, F_OK) == 0) return 1;

#ifndef MESS
	switch (filetype)
	{
		case OSD_FILETYPE_ROM:
			sprintf(name,"%s/%s.zip","roms", newfilename);
			if (access(name, F_OK) == 0) return 1;
			break;
		case OSD_FILETYPE_SAMPLE:
			sprintf(name,"%s/%s.zip","samples", newfilename);
			if (access(name, F_OK) == 0) return 1;
			break;
	}
#endif
	
	return 0;
}

/*
 * Search file caseinsensitively
 *
 * Arguments: 
 *	char * path - complete pathname to the desired file. The string will
 *	              be modified during search (and contains the final output).
 *
 * Return TRUE if found, FALSE otherwise.
 */
int filesearch(char *path)
{
    DIR *dirp;
    struct dirent *de = NULL;
    char *ep, *dp, *fp;

    ep = strrchr(path, '/');
    if (ep) {
	*ep = '\0';	/* I guess root directory is not supported */
	dp = path;
	fp = ep + 1;
    } else {
	dp = "."; /* well, what should be the correct name for "current dir" */
	fp = path;
    }

    if (*fp == '\0') {
	return FALSE;
    }

    /* jamc: code to perform a upper/lowercase rom search */
    /* try to open directory */
    if ((dirp = opendir(dp)) == (DIR *) 0) { 
	return FALSE;
    }

    /* search entry and upcasecompare to desired file name */
    for (de = readdir(dirp); de; de = readdir(dirp))
	if (!strcasecmp(de->d_name, fp)) break;
    if (de) strcpy(fp, de->d_name);
    closedir(dirp);
    
    if (ep) *ep = '/';

    if (de) return TRUE;
    return FALSE;
}

void *osd_fopen(const char *gamename, const char *filename, int filetype, 
 int write) {
	char *pt, name[256]; /* FIXME ?? possible buffer overrun */
	FakeFileHandle *f;

	f = (FakeFileHandle *)malloc(sizeof(FakeFileHandle));
	if (f == NULL) return f;
	
	f->type = kPlainFile;
	f->file = 0;

	switch (filetype)
	{
#ifdef MESS
		case OSD_FILETYPE_ROM_CART:
		case OSD_FILETYPE_IMAGE:
			sprintf(name,"%s",filename);
#else
		case OSD_FILETYPE_ROM:
		case OSD_FILETYPE_SAMPLE:
			sprintf(name,"%s/%s",gamename,filename);
#endif
			if (filesearch(name))
				f->file = fopen(name,write ? "wb" : "rb");
			if (f->file == 0)
			{
				/* try with a .zip extension */
#ifdef MESS
				sprintf(name,"%s.zip", gamename);
#else
				if  (filetype == OSD_FILETYPE_ROM)
					sprintf(name,"%s/%s.zip","roms", gamename);
				else
					sprintf(name,"%s/%s.zip","samples", gamename);
#endif
				f->file = fopen(name, write ? "wb" : "rb");
				if (f->file)
				{
					if (errorlog)
						fprintf(errorlog, "using zip file for %s\n", filename);
					fclose(f->file);
					f->type = kZippedFile;
					if (load_zipped_file(name, filename, &f->data, &f->length))
					{
						f->data = 0;
						f->file = 0;
					}
					else
						f->offset = 0;
				}
			}
			break;
		case OSD_FILETYPE_HIGHSCORE:
			if (mame_highscore_enabled())
			{
				sprintf(name,"%s/%s.hi", spooldir, gamename);
				f->file = fopen(name,write ? "wb" : "rb");
			}
			break;
		case OSD_FILETYPE_CONFIG:
			pt = getenv("HOME");
			if (pt)
			{
#ifdef MESS
				sprintf(name,"%s/%s/%s.cfg",pt,"xmess",gamename);
#else
				sprintf(name,"%s/%s/%s.cfg",pt,"xmame",gamename);
#endif
				f->file = fopen(name,write ? "wb" : "rb");
			}
			break;
		case OSD_FILETYPE_INPUTLOG:
			pt = getenv("HOME");
			if (pt)
			{
				sprintf(name,"%s/%s.inp",pt,filename);
				f->file = fopen(name,write ? "wb" : "rb");
			}
			break;
		default:
			break;
	}

	if (f->file == 0)
	{
		free(f);
		return NULL;
	}

	return f;
}



int osd_fread(void *file,void *buffer,int length)
{
	FakeFileHandle *f = (FakeFileHandle *)file;
	int read = 0;

	switch (f->type)
	{
		case kPlainFile:
			return fread(buffer,1,length,f->file);
			break;
		case kZippedFile:
			/* reading from the uncompressed image of a zipped file */
			if (f->data)
			{
				if (length + f->offset > f->length)
					length = f->length - f->offset;
				memcpy(buffer, f->offset + f->data, length);
				f->offset += length;
				return length;
			}
			break;
	}

	return read;
}



int osd_fwrite(void *file,const void *buffer,int length)
{
	return fwrite(buffer,1,length,((FakeFileHandle *)file)->file);
}



int osd_fseek(void *file,int offset,int whence)
{
	FakeFileHandle *f = (FakeFileHandle *)file;
	int err = 0;

	switch (f->type)
	{
		case kPlainFile:
			return fseek(((FakeFileHandle *)file)->file,offset,whence);
			break;
		case kZippedFile:
			/* seeking within the uncompressed image of a zipped file */
			switch (whence)
			{
				case SEEK_SET:
					f->offset = offset;
					break;
				case SEEK_CUR:
					f->offset += offset;
					break;
				case SEEK_END:
					f->offset = f->length + offset;
					break;
			}
			break;
	}

	return err;
}



void osd_fclose(void *file)
{
	FakeFileHandle *f = (FakeFileHandle *) file;

	switch(f->type)
	{
		case kPlainFile:
			fclose(f->file);
			break;
		case kZippedFile:
			if (f->data)
				free(f->data);
			break;
	}
	free(f);
}
